/**
  CS106B Section 2
  ----------------
**/

#include <iostream>
#include <string>
#include "console.h"
#include "filelib.h"
#include "simpio.h"
#include "set.h"
#include "map.h"
#include "vector.h"
#include "lexicon.h"
#include <cmath>
#include "gwindow.h"

using namespace std;

struct ListNode {
    int data;
    ListNode *next;
};

struct StringNode {
    string data;
    StringNode *next;
};

struct DoubleNode {
    double data;
    DoubleNode *next;
};

struct PointNode {
    int x;
    int y;
    PointNode *next;
};


void seriesOfUnfortunateReferencesTest();
void VFD(int *duncan, int isadora, int& quigley);

void isSortedTest();
bool isSorted(ListNode *front);

void countDuplicatesTest();
int countDuplicateStrings(StringNode *front);

void removeAllThresholdTest();
void removeAllThreshold(DoubleNode *&front, double value, double threshold);

void doubleListTest();
void doubleList(ListNode *&front);

void splitTest();
void split(ListNode *&front);

void reverseTest();
ListNode *reverse(ListNode *front);

void mergeListsTest();
ListNode *mergeLists(ListNode *one, ListNode *two);

void drawPolygonalPathTest();
void drawPolygonalPath(GWindow &window, PointNode *head);

void braidTest();
void braid(ListNode *list);

ListNode *createListNodes(const Vector<int> &v);
void deleteListNodes(ListNode *head);
void deleteStringNodes(StringNode *head);
StringNode *createStringNodes(const Vector<string> &v);
DoubleNode *createDoubleNodes(const Vector<double> &v);
void deleteDoubleNodes(DoubleNode *head);
void printDoubleList(DoubleNode *head);
void printIntList(ListNode *head);
PointNode *createPointNodes(const Vector<int> &x, const Vector<int> &y);
void deletePointNodes(PointNode *head);

int main() {
    seriesOfUnfortunateReferencesTest();
    isSortedTest();
    countDuplicatesTest();
    removeAllThresholdTest();
    doubleListTest();
    splitTest();
    reverseTest();
    mergeListsTest();
    drawPolygonalPathTest();
    braidTest();
    return 0;
}

/** A Series of Unfortunate References **/

void seriesOfUnfortunateReferencesTest() {
    cout << "== A SERIES OF UNFORTUNATE REFERENCS ==" << endl;
    int sunny = -6;
    int klaus = 21;
    int violet = 2;
    VFD(&sunny, violet, klaus);
    cout << sunny << " " << klaus << " " << violet << " " << endl;
}

void VFD(int *duncan, int isadora, int &quigley) {
    isadora += *duncan;
    quigley *= isadora;
    (*duncan) = 1;
    isadora = *duncan;
}

/** Is Sorted **/

void isSortedTest() {
    cout << "== IS SORTED ==" << endl;
    Vector<int> v;
    ListNode *list = createListNodes(v);
    cout << "Sorted? " << isSorted(list) << endl;
    deleteListNodes(list);

    v = {1, 2, 3, 4, 5, 6, 7, 7, 8};
    list = createListNodes(v);
    cout << "Sorted? " << isSorted(list) << endl;
    deleteListNodes(list);

    v = {1, 2, 3, 2, 6, 7};
    list = createListNodes(v);
    cout << "Sorted? " << isSorted(list) << endl;
    deleteListNodes(list);
}

bool isSorted(ListNode *front) {
    if (front != NULL) {
        ListNode *current = front;
        while (current->next != NULL) {
            if (current->data > current->next->data) {
                return false;
            }
            current = current->next;
        }
    }
    return true;
}

/** Count Duplicates **/

void countDuplicatesTest() {
    cout << "== COUNT DUPLICATES ==" << endl;
    Vector<string> v = {"apple", "apple", "apple", "bat", "bat", "car", "car", "dog", "dog", "dog", "fox", "fox"};
    StringNode *head = createStringNodes(v);
    cout << countDuplicateStrings(head) << " duplicates" << endl;
    deleteStringNodes(head);
}

int countDuplicateStrings(StringNode *front) {
    int count = 0;
    if (front != NULL) {
        StringNode *current = front;
        while (current->next != NULL) {
            if (current->data ==current->next->data) {
                count++;
            }
            current = current->next;
        }
    }
    return count;
}

/** Remove all threshold **/
void removeAllThresholdTest() {
    cout << "== Remove All Threshold ==" << endl;
    Vector<double> v = {3.0, 9.0, 4.2, 2.1, 3.3, 2.3, 3.4, 4.0, 2.9, 2.7, 3.1, 18.2};
    DoubleNode *head = createDoubleNodes(v);
    removeAllThreshold(head, 3, 0.3);
    printDoubleList(head);
    deleteDoubleNodes(head);
}

void removeAllThreshold(DoubleNode *&front, double value, double threshold) {
    while (front != NULL && front->data >= value - threshold && front->data <= value + threshold) {
        DoubleNode *trash = front;
        front = front->next;
        delete trash;
    }
    if (front != NULL) {
        DoubleNode *current = front;
        while (current->next != NULL) {
            if (current->next->data >= value - threshold && current->next->data <= value + threshold) {
                DoubleNode *trash = current->next;
                current->next = current->next->next;
                delete trash;
            } else {
                current = current->next;
            }
        }
    }
}

/** Double List **/
void doubleListTest() {
    cout << "== DOUBLE LIST ==" << endl;
    Vector<int> v = {1, 3, 2, 7};
    ListNode *head = createListNodes(v);
    doubleList(head);
    printIntList(head);
    deleteListNodes(head);
}

void doubleList(ListNode *&front) {
    if (front != NULL) {
        ListNode *half2 = new ListNode;
        half2->data = front->data;
        half2->next = NULL;
        ListNode *back = half2;
        ListNode *current = front;
        while (current->next != NULL) {
            current = current->next;
            back->next = new ListNode;
            back->next->data = current->data;
            back->next->next = NULL;
            back = back->next;
        }
        current->next = half2;
    }
}

/** Split **/
void splitTest() {
    cout << "== SPLIT ==" << endl;
    Vector<int> v = {-5, 8, 7, -4, 19, 0, 43, -8, -7, 2};
    ListNode *head = createListNodes(v);
    split(head);
    printIntList(head);
    deleteListNodes(head);
}

void split(ListNode *&front) {
    if (front != nullptr) {
        // Points to the latest negative number found
        ListNode *lastNegative = nullptr;
        if (front->data < 0) {
            lastNegative = front;
        }

        // Operate one element ahead in case we need to rewire
        ListNode *current = front;
        while (current->next != nullptr) {
            if (current->next->data < 0) {
                // Have current element reroute around this element
                ListNode *temp = current->next;
                current->next = current->next->next;
                if (lastNegative == nullptr) {
                    // If first negative we've seen, it should be at front
                    lastNegative = temp;
                    lastNegative->next = front;
                    front = lastNegative;
                } else {
                    // This should be at the end of the negatives list
                    temp->next = lastNegative->next;
                    lastNegative->next = temp;
                    lastNegative = temp;
                    // Jump ahead if this element is still next
                    // (could happen if front is negative)
                    if (current->next == lastNegative) {
                        current = current->next;
                    }
                }
            } else {
                current = current->next;
            }
        }
    }
}

/** Reverse **/
void reverseTest() {
    cout << "== REVERSE ==" << endl;
    Vector<int> v = {1, 2, 3, 4, 5, 8};
    ListNode *head = createListNodes(v);
    head = reverse(head);
    printIntList(head);
    deleteListNodes(head);
}

ListNode *reverse(ListNode *front) {
    if (front == NULL) return NULL;
    if (front->next == NULL) return front;

    ListNode *rest = reverse(front->next);  // reverse the whole list but the front
    front->next->next = front;              // update the next pointer of the 2nd element (now
    // the second to last element) to point to front
    front->next = NULL;                     // update the next pointer of the front (now the
    // end of the list) to be NULL.
    return rest;
}

/** Merge Lists **/
void mergeListsTest() {
    cout << "== MERGE LISTS ==" << endl;
    Vector<int> v1 = {1, 2, 3, 4, 5, 8};
    Vector<int> v2 = {6, 7, 10, 11};
    ListNode *head1 = createListNodes(v1);
    ListNode *head2 = createListNodes(v2);
    ListNode *merged = mergeLists(head1, head2);
    printIntList(merged);
    deleteListNodes(merged);
}

ListNode *mergeLists(ListNode *one, ListNode *two) {
    if (one == NULL) return two;
    if (two == NULL) return one;

    // got this far? then neither list is empty
    if (one->data <= two->data) {
        one->next = mergeLists(one->next, two);
        return one;
    } else {
        two->next = mergeLists(one, two->next);
        return two;
    }
}

/** Draw Polygonal Path **/
void drawPolygonalPathTest() {
    cout << "== DRAW POLYGONAL PATH (see GWindow) ==" << endl;
    GWindow window;
    Vector<int> xs = {0, 100, 200};
    Vector<int> ys = {0, 100, 50};
    PointNode *head = createPointNodes(xs, ys);
    drawPolygonalPath(window, head);
    deletePointNodes(head);
}

void drawPolygonalPath(GWindow &window, PointNode *head) {
    if (head == NULL) {
        return;
    }
    PointNode *current = head;
    window.fillOval(current->x - 1, current->y - 1, 2, 2);
    while (current->next != NULL) {
        window.drawLine(current->x, current->y, current->next->x, current->next->y);
        window.fillOval(current->next->x - 1, current->next->y - 1, 2, 2);
        current = current->next;
        if (current == head) {
            break;
        }
    }
}

/** Braid **/
void braidTest() {
    cout << "== BRAID ==" << endl;
    Vector<int> v = {1, 4, 2};
    ListNode *head = createListNodes(v);
    braid(head);
    printIntList(head);
    deleteListNodes(head);
}

void braid(ListNode *list) {
    ListNode *reverse = NULL;
    for (ListNode *curr = list; curr != NULL; curr = curr->next) {
        ListNode *newNode = new ListNode;
        newNode->data = curr->data;
        newNode->next = reverse;
        reverse = newNode;
    }

    // reverse now addresses a memory-independent
    // version of the original list,
    // where all of the nodes are in reverse order.
    ListNode *rest = reverse;   // rest addresses part
                                // that has yet to be braided
    for (ListNode *curr = list; curr != NULL; curr = curr->next->next) {
        ListNode *next = rest->next;
        rest->next = curr->next;
        curr->next = rest;
        rest = next;
    }
}

/** Utility functions **/

ListNode *createListNodes(const Vector<int>& v) {
    if (v.isEmpty()) {
        return NULL;
    }

    ListNode *head = new ListNode;
    head->data = v[0];
    head->next = nullptr;

    ListNode *curr = head;
    for (int i = 1; i < v.size(); i++) {
        ListNode *newNode = new ListNode;
        newNode->data = v[i];
        newNode->next = nullptr;
        curr->next = newNode;
        curr = newNode;
    }

    return head;
}

void deleteListNodes(ListNode *head) {
    if (head == NULL) return;
    if (head->next != NULL) {
        deleteListNodes(head->next);
    }

    delete head;
}

StringNode *createStringNodes(const Vector<string>& v) {
    if (v.isEmpty()) {
        return NULL;
    }

    StringNode *head = new StringNode;
    head->data = v[0];
    head->next = nullptr;

    StringNode *curr = head;
    for (int i = 1; i < v.size(); i++) {
        StringNode *newNode = new StringNode;
        newNode->data = v[i];
        newNode->next = nullptr;
        curr->next = newNode;
        curr = newNode;
    }

    return head;
}

void deleteStringNodes(StringNode *head) {
    if (head == NULL) return;
    if (head->next != NULL) {
        deleteStringNodes(head->next);
    }

    delete head;
}

DoubleNode *createDoubleNodes(const Vector<double> &v) {
    if (v.isEmpty()) {
        return NULL;
    }

    DoubleNode *head = new DoubleNode;
    head->data = v[0];
    head->next = nullptr;

    DoubleNode *curr = head;
    for (int i = 1; i < v.size(); i++) {
        DoubleNode *newNode = new DoubleNode;
        newNode->data = v[i];
        newNode->next = nullptr;
        curr->next = newNode;
        curr = newNode;
    }

    return head;
}

void deleteDoubleNodes(DoubleNode *head) {
    if (head == NULL) return;
    if (head->next != NULL) {
        deleteDoubleNodes(head->next);
    }

    delete head;
}

void printDoubleList(DoubleNode *head) {
    if (head == NULL) return;
    if (head->next == NULL) {
        cout << "[" << head->data << "]" << endl;
    } else {
        cout << "[" << head->data << "] -> ";
    }

    printDoubleList(head->next);
}

void printIntList(ListNode *head) {
    if (head == NULL) return;
    if (head->next == NULL) {
        cout << "[" << head->data << "]" << endl;
    } else {
        cout << "[" << head->data << "] -> ";
    }

    printIntList(head->next);
}

PointNode *createPointNodes(const Vector<int> &x, const Vector<int> &y) {
    assert(x.size() == y.size());

    if (x.isEmpty()) {
        return NULL;
    }

    PointNode *head = new PointNode;
    head->x = x[0];
    head->y = y[0];
    head->next = nullptr;

    PointNode *curr = head;
    for (int i = 1; i < x.size(); i++) {
        PointNode *newNode = new PointNode;
        newNode->x = x[i];
        newNode->y = y[i];
        newNode->next = nullptr;
        curr->next = newNode;
        curr = newNode;
    }

    return head;
}

void deletePointNodes(PointNode *head) {
    if (head == NULL) return;
    if (head->next != NULL) {
        deletePointNodes(head->next);
    }

    delete head;
}
